
use v5.14;
use warnings;

package state51::Feature {
	use JSON::MaybeXS qw[ decode_json ];
	use LWP::UserAgent;
	use HTTP::Request::Common qw[ GET ];

	use namespace::clean;

	our $URI = $ENV{FEATURE_SWITCH_URI};
	our $INTERVAL = 5 * 60; # refresh configuration every 5 minutes

	sub is_enabled {
		my ($class, $feature) = @_;

		state $next_refresh = 0;
		state $known_features;

		if (time > $next_refresh || ! $known_features) {
			state $ua = LWP::UserAgent->new;

			$known_features = {};

			$next_refresh = time + $INTERVAL;
			$next_refresh -= $next_refresh % $INTERVAL;

			# returns all features with their sample value and status
			# not only enabled ones
			my $result = $ua->request (GET $URI, Accept => 'application/json');

			if ($result->is_success) {
				$known_features = {
						map { $_->{name} => { sample => $_->{sample}, status => $_->{status} } }
						@{ decode_json ($result->decoded_content) }
				};
			}
		}

		# can't be enabled if it doesn't exist or the service is not available
		if (not exists $known_features->{$feature}) {
			return 0;
		}

		# always return 0 if the feature is disabled
		return 0 if $known_features->{$feature}->{status} == 0;

		# return 1 all the time if sample is 100
		# and some of the time if sample is less than 100
		if (rand(100) < $known_features->{$feature}->{sample}) {
                    warn "feature $feature is enabled.";
                    return 1;
                }
	}

	1;
}

__END__

=pod

=head1 NAME

state51::Feature - Feature switch access library

=head2 SYNOPSIS

	use state51::Feature

	if (state51::Feature-><feature_name>) {
		...
	}

=head2 DESCRIPTION

Module provides interface to usage of features.
Enabled/disabled features are configured in feature switch service,
accessible via env variable C<FEATURE_SWITCH_URI>.

Every feature is exposed as a method returning true if feature is
in enabled state.

Module warns when feature is not defined and returns false then.

When feature service is not available it always returns false;

Module refreshes feature configuration every 5 minutes.

=cut
