package state51::RabbitMQ;

use MooseX::Singleton;
use MooseX::Types::Moose qw/ Object /;
use AnyEvent::RabbitMQ;
use File::ShareDir ();
use Try::Tiny;
use state51::AnyEvent qw/ FULL_ASYNC /;
use Scalar::Util qw/ blessed /;
use Data::Dumper;

my $spec = File::ShareDir::dist_dir("AnyEvent-RabbitMQ") . '/fixed_amqp0-9-1.xml';
my $rf;

sub _make_rf {
    if (!$rf) {
        $rf = AnyEvent::RabbitMQ->new(
            verbose => $ENV{RABBITMQ_DEBUG} || 0,
        )->load_xml_spec($spec);
    }
    $rf;
}

sub host { $ENV{PLATFORM_MQ_SERVER} || 'localhost' }
sub port { $ENV{PLATFORM_MQ_PORT} || 5672 }

my $RECONNECT_OK = 1;

my ($mq, $mq_pid);
my @_mq_connected_cbs;
our $CONNECTING;
sub _mq_connected {
    undef $CONNECTING;
    my @cbs = @_mq_connected_cbs;
    @_mq_connected_cbs = ();
    $_->($mq) for @cbs;
    return $mq;
}

sub _do_reconnect_on_idle {
    return unless $RECONNECT_OK;
    my $idle; $idle = AnyEvent->idle(cb => sub { undef $idle; __PACKAGE__->instance->mq });
}

sub _reconnect {
    my $self = shift;
    return unless $RECONNECT_OK;
    my $cv = AnyEvent->condvar;
    my $fail = sub {
        undef $mq;
        undef $CONNECTING;
        $cv->send if blessed($cv);
        return unless $RECONNECT_OK;
        warn(sprintf("Reconnection to RabbitMQ on %s:%s failed!\n" , $self->host, $self->port));
    };
    _make_rf()->connect(
        on_success => sub {
            undef $CONNECTING;
            $mq = $rf;
            $mq_pid = $$;
            if (blessed($cv)) {
                $cv->send;
            }
            _mq_connected();
        },
        on_close => $fail,
        on_failure => $fail,
        on_read_failure => $fail,
        host => __PACKAGE__->instance->host,
        port => __PACKAGE__->instance->port,
        user => $ENV{PLATFORM_MQ_USER} || $fail,
        pass => $ENV{PLATFORM_MQ_PASSWORD} || $fail,
        vhost => $ENV{PLATFORM_MQ_VHOST} || '/',
        timeout => 2,
        tune => { heartbeat => 30 },
    );
}

my $cv;
sub mq {
    my ($self, $cb) = @_;
    push(@_mq_connected_cbs, $cb) if $cb;
    if ($mq) {
        $cb->($mq) if $cb;
        return $mq;
    }
    return if $CONNECTING;
    return unless $RECONNECT_OK;
    $cv = AnyEvent->condvar;
    $CONNECTING = 1;
    AnyEvent->now_update;
    my $timer; $timer = AnyEvent->timer(after => 5, cb => sub {
        warn(sprintf("Connection to RabbitMQ on %s:%s timed out!\n" , $self->host, $self->port));
        $cv->send if blessed($cv);
        undef $rf;
        undef $mq;
        undef $CONNECTING;
        $self->_reconnect();
    });
    my $close = sub {
        undef $mq;
        undef $timer;
        undef $CONNECTING;
        $cv->send if blessed($cv);
        return unless $RECONNECT_OK;
        warn(sprintf("Connection to RabbitMQ on %s:%s closed: %s\n" , $self->host, $self->port, join(', ', @_)));
        $self->_reconnect();
    };
    my $fail = sub {
        undef $mq;
        undef $timer;
        undef $CONNECTING;
        $cv->send if blessed($cv);
        return unless $RECONNECT_OK;
        warn(sprintf("Connection to RabbitMQ on %s:%s failed: %s\n" , $self->host, $self->port, join(', ', @_)));
        $self->_reconnect();
    };
    my $read_fail = sub {
        undef $mq;
        undef $timer;
        undef $CONNECTING;
        $cv->send if blessed($cv);
        return unless $RECONNECT_OK;
        warn(sprintf("Connection to RabbitMQ on %s:%s read failure: %s\n" , $self->host, $self->port, join(', ', @_)));
        $self->_reconnect();
    };
    _make_rf()->connect(
        on_success => sub {
            undef $timer;
            undef $CONNECTING;
            $mq = $rf;
            $mq_pid = $$;
            if (blessed($cv)) {
                $cv->send;
            }
            _mq_connected();
        },
        on_close => $close,
        on_failure => $fail,
        on_read_failure => $read_fail,
        host => __PACKAGE__->instance->host,
        port => __PACKAGE__->instance->port,
        user => $ENV{PLATFORM_MQ_USER} || $fail,
        pass => $ENV{PLATFORM_MQ_PASSWORD} || $fail,
        vhost => $ENV{PLATFORM_MQ_VHOST} || '/',
        timeout => 2,
        tune => { heartbeat => 30 },
    );
    $cv->recv unless FULL_ASYNC;
    undef $cv;
    $timer = undef;
    return $mq;
}

END {
    $RECONNECT_OK = 0;
    if ($mq and $mq_pid == $$) {
        try {
            $cv = AnyEvent->condvar;
            $mq->close(
                on_success => sub { $cv->send if blessed($cv); },
                on_failure => sub { $cv->croak(shift()) if $cv },
            );
            AnyEvent->now_update;
            my $t; $t = AnyEvent->timer(after => 2, cb => sub { undef $t; $cv->croak("Timed out closing connection after 2s") });
            $cv->recv;
            undef $cv;
            undef $t;
        }
        catch {
            warn("MQ close failed: $_");
        };
    }
}

__PACKAGE__->meta->make_immutable();
1;

