package state51::Script::MessageJobs::Worker;
use state51::Moose;
use state51::RabbitMQ;
use AnyEvent;
use JSON;
use MooseX::Types::Moose qw/ Str ArrayRef Bool Int /;
use List::Util qw/ shuffle /;
use List::MoreUtils qw/ uniq /;
use Data::Dumper;
use FindBin;
use Scalar::Util qw( weaken );

use state51::Logger;
use state51::Logger::Event::JobFailedDeserialise;

has cooldown_type => (
    is => 'ro',
    isa => Str,
    default => 'workerpool'
);

has heartbeat_file => (
    is => 'ro',
    isa => Str,
    predicate => '_has_heartbeat_file',
);

with 'state51::Script::Role::WorkPool::Worker';

around log_message => sub {
    my ($orig, $self, $message, $level) = @_;
    my $queues = join(", ", @{ $self->queue_names });
    $self->$orig($queues . ": " . $message, $level);
};

has queue_names => (
    isa => ArrayRef[Str],
    is => 'ro',
    required => 1,
);

has _priority_levels => (
    is => 'ro',
    isa => ArrayRef[Int],
    lazy => 1,
    default => sub {
        my ($self) = @_;
        return [
            sort { $a <=> $b }
            uniq
            map { /^(\d+):/; $1; }
            grep { /^\d+:/ }
            @{ $self->queue_names }
        ];
    },
);

around BUILDARGS => sub {
    my $orig = shift;
    my $class = shift;
    my %args = ref($_[0]) ? %{ $_[0] } : @_;

    if ($args{queue_name}) {
        $args{queue_names} = [ delete($args{queue_name}) ];
    }
    return $class->$orig(%args);
};


sub get_job {
    my $self = shift;

    foreach my $priority (@{$self->_priority_levels}) {
        foreach my $queue_name (shuffle grep { $_ =~ /^$priority:/ }
            @{$self->queue_names})
        {
            $queue_name =~ /^\d+:(.*)/ or die;
            my $message = $self->_get_job_from_queue($1);
            my ($job, $correlation_id, $reply_to);

            if (defined $message) {
                $correlation_id = $message->{header}->correlation_id;
                $reply_to = $message->{header}->reply_to;
                $job = $message->{body}->payload;
            }
            return [$job, $correlation_id, $reply_to] if defined $job;
        }
    }

    # Test queues with no numeric prefixs
    foreach
      my $queue_name (shuffle grep { $_ !~ /^\d+:/ } @{$self->queue_names})
    {
        my $message = $self->_get_job_from_queue($queue_name);
        my ($job, $correlation_id, $reply_to);

        if (defined $message) {
            $correlation_id = $message->{header}->correlation_id;
            $reply_to = $message->{header}->reply_to;
            $job = $message->{body}->payload;
        }
        return [$job, $correlation_id, $reply_to] if defined $job;
    }

    return;
}

sub _get_job_from_queue {
    my ($self, $queue_name) = @_;

    my $cv = AnyEvent->condvar;
    state51::RabbitMQ->instance->mq(sub {
        my ($mq) = @_;
        $mq->open_channel(
            on_success => sub {
                my ($channel) = @_;
                {
                weaken($channel);

                $channel->get(
                    queue => $queue_name,
                    on_success => sub {
                        $channel->close();
                        my $message = shift;
                        if ($message->{ok}) {
                            $cv->send($message);
                        }
                        else {
                            $cv->send(undef);
                            return if $message->{empty};
                            warn Data::Dumper::Dumper($message);
                        }
                    },
                    on_failure => sub {
                        Carp::cluck("Getting message from " . $queue_name . " unexpectedly failed: " . Data::Dumper::Dumper(@_));
                        $cv->send(undef);
                    },
                );
                }
            },
            on_failure => sub { $cv->croak(Dumper(@_)) },
            on_close => sub { $cv->croak(Dumper(@_)) },
        );
    });
    return $cv->recv;
}

sub do_job {
    my ($self, $job, $correlation_id, $reply_to) = @_;
    my $data = JSON::decode_json($job);

    # Set up heartbeat timer for long-running jobs
    my $heartbeat_timer;
    if ($self->_has_heartbeat_file && -e $self->heartbeat_file) {
        $heartbeat_timer = AnyEvent->timer(
            after    => 300,  # First heartbeat after 5 minutes
            interval => 300,  # Heartbeat every 5 minutes after that
            cb       => sub {
                # Update the heartbeat file with current timestamp
                open my $fh, '>', $self->heartbeat_file or warn "Cannot open heartbeat file: $!";
                print $fh time;
                close $fh;
                $self->log_message("Sent heartbeat for long-running job", "debug");
            }
        );
    }

    # Sneakily add the correlation id from the envelope to the message
    if ($correlation_id) {
        $data->{correlation_id} = $correlation_id;
    }

    my $object;
    eval {
        Class::Load::load_class($data->{__CLASS__});
        $object = $data->{__CLASS__}->unpack($data);
    };
    if ($@) {
        state51::Logger->new->message(
            level   => 'info',
            message => [
                state51::Logger::Event::GreedbagCheckout::JobFailedDeserialise
                  ->new(
                    job_data  => $data,
                    error => $@,
                    ApplicationID => '54838309120320',
                    Environment => 'production',
                  )
            ]
        );
        warn "Failed to unpack message: $@";
        return;
    }


    local $0 = sprintf(
        "%s %s: %s%s",
        "perl", # $^X
        $FindBin::Script,
        ref($object),
        $object->can("command_line") ? $object->command_line() : "",
    );

    # Update heartbeat file immediately before starting job
    if ($self->_has_heartbeat_file) {
        open my $fh, '>', $self->heartbeat_file or warn "Cannot open heartbeat file: $!";
        print $fh time;
        close $fh;
    }

    $object->run;

    # Heartbeat timer will be automatically destroyed when it falls out of scope
}

__PACKAGE__->meta->make_immutable;
1;

