package state51::Script::MessageJobs::WorkerPool;
use state51::Moose;
use MooseX::Types::Moose qw/ Str ArrayRef Int /;
use state51::RabbitMQ;
use File::Temp;

has cooldown_type => (
    is      => 'ro',
    isa     => Str,
    default => 'workerpool'
);

with 'state51::Script::Role::WorkPool::Runner';

around log_message => sub {
    my ( $orig, $self, $message, $level ) = @_;
    my $queues = join( ", ", @{ $self->queue_names } );
    $self->$orig( "$queues: " . $message, $level );
};

has queue_names => (
    isa      => ArrayRef [Str],
    is       => 'ro',
    required => 1,
);

has '+worker_class' => ( default => 'state51::Script::MessageJobs::Worker' );

has '+worker_options' =>
  ( default => sub { [qw/number_of_jobs_before_terminate queue_names heartbeat_file /] } );

has heartbeat_file => (
    is      => 'ro',
    isa     => Str,
    lazy    => 1,
    default => sub {
        my $self = shift;
        return File::Temp::tmpnam() . '.heartbeat';
    },
);

has _timer => (
    is      => 'ro',
    lazy    => 1,
    default => sub {
        my $self = shift;
        AnyEvent->timer(
            after    => 30,
            interval => 30,
            cb       => sub {
                $self->setup_watcher unless $self->_has_channel;

                # Check for heartbeat file updates from workers
                if (-e $self->heartbeat_file) {
                    my $mtime = (stat($self->heartbeat_file))[9];
                    if ($mtime > $self->_last_consumed) {
                        $self->_last_consumed($mtime);
                    }
                }

                if ( $self->_last_consumed < ( time - 3600 ) ) {
                    warn "No jobs seen since "
                      . $self->_last_consumed
                      . " stopping";
                    $self->stop(1);
                }
            }
        );
    },
);

has _channel => (
    is        => 'rw',
    clearer   => '_clear_channel',
    predicate => '_has_channel',
);

has _last_consumed => (
    is      => 'rw',
    isa     => Int,
    lazy    => 1,
    default => sub { time },
);

sub additional_monitoring_data {
    my ($self) = @_;
    my %data = ( time_since_last_consume => ( time - $self->_last_consumed ), );

    return %data;
}

around BUILDARGS => sub {
    my $orig  = shift;
    my $class = shift;
    my %args  = ref( $_[0] ) ? %{ $_[0] } : @_;

    if ( $args{queue_name} ) {
        $args{queue_names} = [ delete( $args{queue_name} ) ];
    }
    return $class->$orig(%args);
};

sub BUILD {
    my $self = shift;
    # Create the heartbeat file
    open my $fh, '>', $self->heartbeat_file or warn "Cannot create heartbeat file: $!";
    print $fh time;
    close $fh;

    $self->_timer;
    $self->setup_watcher;
}

sub setup_watcher {
    my $self = shift;

    state51::RabbitMQ->instance->mq(
        sub {
            my $mq = shift;
            $mq->open_channel(
                on_success => sub {
                    my $channel = shift;
                    warn "Connecting to RabbitMQ";

                    $self->_channel($channel);

                    $channel->declare_queue(
                        auto_delete => 1,
                        exclusive   => 1,
                        on_success  => sub {
                            my $anon_queue = shift()->method_frame->queue;
                            my $bindings   = 0;
                            foreach my $queue_name_with_priority (
                                @{ $self->queue_names } )
                            {
                                my $queue_name = $queue_name_with_priority;
                                $queue_name =~ s/^\d+://;

                                $channel->declare_exchange(
                                    type       => 'topic',
                                    exchange   => $queue_name,
                                    durable    => 1,
                                    on_success => sub {
                                        $channel->declare_queue(
                                            queue      => $queue_name,
                                            durable    => 1,
                                            on_success => sub {
                                                $channel->bind_queue(
                                                    queue       => $queue_name,
                                                    exchange    => $queue_name,
                                                    routing_key => '#',
                                                    on_success  => sub {
                                                        $channel->bind_queue(
                                                            queue =>
                                                              $anon_queue,
                                                            exchange =>
                                                              $queue_name,
                                                            routing_key => '#',
                                                            on_success  => sub {
                                                                $bindings++;
                                                                if (
                                                                    $bindings
                                                                    == scalar(
                                                                        @{
                                                                            $self
                                                                              ->
                                                                              queue_names
                                                                        }
                                                                    )
                                                                  )
                                                                {
                                                                    warn
"All queues bound, consuming";
                                                                    $channel
                                                                      ->consume(
                                                                        queue =>
                                                                          $anon_queue,
                                                                        on_success =>
                                                                          sub {
                                                                            warn
"Watching anonymous queue";
                                                                          },
                                                                        on_failure =>
                                                                          sub {
                                                                            Carp::cluck(
"Getting subscribed to anonymous unexpectedly failed: "
                                                                                  .
                                                                                  Data::Dumper::Dumper(
                                                                                    @_
                                                                                  )
                                                                            );
                                                                            $self
                                                                              ->
                                                                              _clear_channel;
                                                                          },
                                                                        on_consume =>
                                                                          sub {
                                                                            $self
                                                                              ->
                                                                              more_to_do
                                                                              (1
                                                                              );
                                                                            $self
                                                                              ->
                                                                              _worker_starter;
                                                                            $self
                                                                              ->
                                                                              _last_consumed
                                                                              (
                                                                                time
                                                                              );
                                                                          },
                                                                      );
                                                                }
                                                            },
                                                            on_failure => sub {
                                                                Carp::cluck(
"Binding exchange to persistent queue failed"
                                                                );
                                                                $self
                                                                  ->_clear_channel;
                                                            },

                                                        );
                                                    },
                                                    on_failure => sub {
                                                        Carp::cluck(
"Binding exchange to persistent queue failed"
                                                        );
                                                        $self->_clear_channel;
                                                    },
                                                );
                                            },
                                            on_failure => sub {
                                                Carp::cluck(
"Declaring persistent queue failed"
                                                );
                                                $self->_clear_channel;
                                            },
                                        );
                                    },
                                    on_failure => sub { $self->_clear_channel },
                                    on_close   => sub { $self->_clear_channel },
                                );
                            }
                        },
                        on_failure => sub {
                            warn Carp::cluck("Making anonymous queue failed");
                            $self->_clear_channel;
                        },
                    );

                },
                on_close   => sub { $self->_clear_channel },
                on_failure => sub { $self->_clear_channel },
            );
        }
    );
}

__PACKAGE__->meta->make_immutable;

