package state51::Script::Role::WorkPool::Runner;
use AnyEvent;
use state51::Moose::Role;
use MooseX::Types::Moose qw/ Bool Int Str ArrayRef HashRef Object /;
use MooseX::Types::LoadableClass qw/ LoadableClass /;
use File::Temp qw/ tempfile /;
use Fcntl qw/ O_APPEND /;
use JSON qw/ encode_json /;
use List::Util qw/ min max /;
use state51::Shared;

use state51::Shared::ForkExecPerl qw/ safe_filename run_perl /;

requires 'additional_monitoring_data';

with qw/
  state51::Script::Role::ScriptSimpleLogger
  state51::Mixin::CanBeCooled
  /;

around log_message => sub {
    my ( $orig, $self, $message, $level ) = @_;
    $self->$orig( "[parent $$] " . $message, $level );
};

has number_of_jobs_before_terminate => (
    is      => 'rw',
    isa     => Int,
    default => 30,
);

has exit_when_complete => (
    is            => 'rw',
    isa           => Bool,
    default       => 1,
    documentation => "Terminate when there is no more work in the queue",
);

has initial_max_workers => (
    isa     => Int,
    is      => 'ro',
    default => sub {
        my ($self) = @_;
        return $self->max_workers() || 1;
    },
    lazy => 1,
);

has max_workers => (
    isa           => Int,
    is            => 'ro',
    default       => 10,
    documentation => "The maximum number of worker processes that will be run",
);

has current_max_workers => (
    isa     => Int,
    is      => 'rw',
    lazy    => 1,
    default => sub { $_[0]->initial_max_workers },
);

has pids => (
    isa     => HashRef [Int],
    default => sub { {} },
    traits  => ['Hash'],
    handles => {
        _setup_pid      => 'set',
        _pid_starttime  => 'get',
        _num_child_pids => 'count',
        _clear_pid      => 'delete',
        _list_pids      => 'keys',
    },
);

has stale_pids => (
    isa     => HashRef [Int],
    default => sub { {} },
    traits  => ['Hash'],
    handles => {
        _mark_pid_stale   => 'set',
        _pid_went_stale   => 'get',
        _count_stale_pids => 'count',
        _clear_stale_pid  => 'delete',
        _list_stale_pids  => 'keys',
    },
);

has pidfile => (
    isa           => Str,
    is            => 'ro',
    documentation => "I can write a pidfile, where do you want it?",
);

has more_to_do => (
    isa     => Bool,
    is      => 'rw',
    default => 1,
    trigger => sub {
        my ( $self, $val ) = @_;
        $self->_worker_starter if $val;
    },
);

has stop => (
    isa     => Bool,
    is      => 'rw',
    default => 0,
    trigger => sub {
        my ( $self, $val ) = @_;
        return unless $val;
        $self->_clear_worker_starter;
        $self->_infanticide;
        $self->_cv_send;
        my $suicide_after = 600;    # 10m for clean shutdown?
        my $suicide_pact;
        $suicide_pact = AnyEvent->timer(
            after => $suicide_after,
            cb    => sub {
                warn(
"Waited $suicide_after seconds for clean shutdown, giving up!"
                );
                $suicide_pact = $suicide_pact;
                exit(1);
            }
        );
    },
);

has _worker_starter => (
    is        => 'ro',
    lazy      => 1,
    builder   => '_build_worker_starter',
    clearer   => '_clear_worker_starter',
    predicate => '_has_worker_starter',
);

has _cv => (
    is      => 'ro',
    lazy    => 1,
    default => sub { AnyEvent->condvar },
    handles => {
        _cv_send => 'send',
    },
    predicate => '_has_cv',
);

has worker_class => (
    isa      => LoadableClass,
    is       => 'ro',
    required => 1,
    coerce   => 1,
);

has worker_options => (
    isa     => ArrayRef [Str],
    is      => 'ro',
    default => sub { ['number_of_jobs_before_terminate'] },
);

has keep_running_file => (
    isa     => Str,
    is      => 'ro',
    default => sub {
        my ( $fh, $fn ) = tempfile( UNLINK => 0, TMPDIR => 1 );
        print $fh "Keep running file for $$\n";
        close $fh;
        return safe_filename($fn);
    },
    lazy => 1,
);

has _update_current_workers => (
    is      => 'ro',
    builder => '_build_update_current_workers'
);

sub _fork_exec {
    my ($self) = @_;

    my @opts;
    foreach my $opt ( 'keep_running_file', @{ $self->worker_options } ) {
        my $vals = $self->$opt();
        foreach my $val ( $vals->flatten ) {
            push( @opts, "--$opt", $val );
        }
    }
    return run_perl( $self->worker_class,
        $self->worker_class . "->new_with_options->run", @opts, );
}

sub _build_update_current_workers {
    my ($self) = @_;
    return AnyEvent->timer(
        after    => 60,
        interval => 60,
        cb       => sub {
            foreach my $pid ( $self->_list_pids ) {
                if ( open( my $fh, "<", "/proc/$pid/cmdline" ) ) {
                    my $line = <$fh>;
                    chomp $line;
                    if ( $line eq '-e' ) {
                        if ( !$self->_pid_went_stale($pid) ) {
                            $self->_mark_pid_stale( $pid, time );
                        }
                    }
                    else {
                        $self->_clear_stale_pid($pid);
                    }

                }
                else {
                    $self->_clear_stale_pid($pid);
                    $self->log_message(
                        "Could not check pid $pid has it already gone? $!",
                        "warn" );
                }
            }
            foreach my $pid ( $self->_list_stale_pids ) {
                if ( time - $self->_pid_went_stale($pid) > 120 ) {
                    $self->log_message(
                        "Sending a SIGTERM to stale process $pid", "trace" );
                    kill( 15, $pid );
                }
            }
        },
    );
}

my $_retry_timer;

sub _build_worker_starter {
    my $self = shift;

    AnyEvent->idle(
        cb => sub {
            sysopen( my $touch_fh, $self->keep_running_file, O_APPEND )
              or die "could not re-open keep alive file: $!";
            utime undef, undef, $touch_fh;    # what does this function return?
            close $touch_fh;

            if ( $self->stop ) {
                $self->log_message( "Stop", "info" );
                $self->_clear_worker_starter;
                return;
            }

            my $cooldown = $self->in_cooldown;

            my $workers_idle =
              ( $self->_num_child_pids < $self->current_max_workers );

            if ( $workers_idle && !$cooldown ) {
                if ( $self->more_to_do ) {
                    my $pid = $self->_fork_exec();
                    $self->_setup_pid( $pid, time() );
                    my $w;
                    $w = AnyEvent->child(
                        pid => $pid,
                        cb  => sub {
                            my ( $pid, $rc ) = @_;
                            $self->_pid_finished( $pid, $rc );
                            $w = undef;
                        }
                    );

                    if ( !kill 0, $pid ) {
                        $self->log_message("Worker $pid died very quickly");
                        $self->_clear_pid($pid);
                        $self->_clear_stale_pid($pid);
                        $w = undef;
                    }
                }
                else {
                    if ( $self->exit_when_complete ) {
                        $self->log_message( "Exiting as complete", "info" );
                        $self->stop(1);
                    }
                    else {
                        $self->_clear_worker_starter;
                        my $t;
                        $t = AnyEvent->timer(
                            after => 60,
                            cb    => sub { $t = undef; $self->_worker_starter; }
                        );
                        $self->more_to_do(1);
                    }
                }
            }
            else {
                my $retry_time = 60;
                if ($cooldown) {
                    $self->log_message( "In cooldown, retrying in $retry_time",
                        "info" );
                }
                $_retry_timer = AnyEvent->timer(
                    after => $retry_time,
                    cb    => sub {
                        $_retry_timer = undef;
                        $self->log_message( "Timed retry", "info" );
                        $self->_worker_starter;
                    }
                );
                $self->_clear_worker_starter;
            }
        }
    );
}

sub _pid_finished {
    my ( $self, $pid, $rc ) = @_;

    $rc = $rc >> 8;
    $self->log_message( "$pid has exited with $rc", "trace" );

    $self->_clear_pid($pid);
    $self->_clear_stale_pid($pid);

    if ( $rc == 0 ) {
        $self->more_to_do(1);
    }
    elsif ( $rc == 1 ) {
        $self->more_to_do(0);
    }
    else {
        $self->log_message(
            "$pid was not happy, exited from child handler with code: $rc",
            "warn" );
        $self->log_message( "shutting down the running cleanly", "info" );
        $self->stop(1);
    }

    return;
}

sub run {
    my ($self) = @_;

    if ( $self->pidfile ) {
        open( my $pidfile, ">", $self->pidfile ) or die $!;
        print $pidfile "$$\n";
        close $pidfile;
    }

    foreach my $sig (qw/ TERM INT /) {
        my $w;
        $w = AnyEvent->signal(
            signal => $sig,
            cb     => sub {
                $self->stop(1);
                $w = undef;
            }
        );
    }

    $self->_worker_starter;
    $self->_cv->recv;
}

sub _infanticide {
    my ($self) = @_;

    warn "removing " . $self->keep_running_file;
    unlink( $self->keep_running_file )
      or warn "Could not remove keep running file: $!";

    # We could also send a SIGTERM with faster results under Linux, but we
    # don't to avoid having more single platform code.
    return;
}

1;

