package state51::Script::Role::WorkPool::Worker;
use state51::Moose::Role;
use Try::Tiny;
use MooseX::Types::Moose qw/ Bool Int Str /;
use File::Temp qw/ tempfile /;
use Sys::Hostname;
if ($ENV{STAY_ALIVE}) {
  require state51::Base;
}

with qw/ state51::Script::Role::ScriptSimpleLogger /;

with qw/ state51::Mixin::CanBeCooled /;

around log_message => sub {
    my ( $orig, $self, $message, $level ) = @_;
    $self->$orig( "[child $$] " . $message, $level );
};

has number_of_jobs_before_terminate => (
    is      => 'ro',
    isa     => Int,
    default => 30,
    traits  => ['Counter'],
    handles => {
        _dec_number_of_jobs_before_terminate => 'dec',
    },
);

has number_of_jobs_run => (
    is      => 'ro',
    isa     => Int,
    default => 0,
    traits  => ['Counter'],
    handles => {
        _inc_number_of_jobs_run => 'inc',
    },
);

requires qw/
  get_job
  do_job
  /;

has verbose => (
    is      => 'ro',
    isa     => Bool,
    default => 1,
);

before do_job => sub {
    my ( $self, $job ) = @_;
    $self->log_message( "running job $job", "trace" ) if $self->verbose;
};

has keep_running_file => (
    is  => 'ro',
    isa => Str,
);

sub run {
    my ($self) = @_;

    my $stop;

    if ($ENV{STAY_ALIVE}) {
        state51::Base->connect();
    }

    local $SIG{TERM} = sub {
        $stop = 1;
        $self->log_message( "SIGTERM received, not taking any more", "trace" );
    };

    my $nothing_to_do;

    do {
        my $response = try { alarm(60); $self->get_job }
        catch { warn "Exception getting job: $_, worker $$ exiting"; exit 0 };
        alarm(0);

        my ($job, $correlation_id, $reply_to);
        if ($response) {
           ($job, $correlation_id, $reply_to) = @$response;
        }

        unless ($job) {
            #$self->log_message( "no pending jobs for $$, giving up", "trace" );
            $stop = 1;
        }
        else {
            $self->_dec_number_of_jobs_before_terminate;
            $stop = 1 if $self->number_of_jobs_before_terminate <= 0;

            $self->_inc_number_of_jobs_run;

            try {
                $self->do_job($job, $correlation_id, $reply_to);
            }
            catch {
                $self->log_message( "Job worker $$ threw an exception: $_",
                    "error" );
                $stop = 1;
            };
        }
        if ( $self->keep_running_file ) {
            if ( !-e $self->keep_running_file ) {
                $self->log_message( "Keep alive file gone, $$ stopping",
                    "trace" );
                $stop = 1;
            }
        }

        if ( kill( 0, getppid() ) != 1 ) {
            $self->log_message(
                "Worker pool process no longer running, $$ stopping", "info" );
            $stop = 1;
        }
        if ($ENV{STAY_ALIVE}) {
            if ($stop) {
                sleep(1);
            }
        }
    } while (($ENV{STAY_ALIVE} || 0) || (!$stop && !$self->in_cooldown ));

    $self->log_message( "worker $$ exiting cleanly.", "trace" );

    $self->_exit( $self->number_of_jobs_run == 0 ? 1 : 0 );
}

sub _exit {
    my ( $self, $code ) = @_;
    exit $code;
}

1;

