package state51::Types;
use MooseX::Types::Moose qw/Int Str Bool ArrayRef Maybe Undef/;
use MooseX::Types::Structured qw(Dict Optional);
use MooseX::Types
    -declare => [qw(
        Barcode
        ISBN10
        DPID
        ISNI
        BlankStr
        CurrencyName
        MaybeBlankCurrencyName
        FileStatus
        ISRC
        MaybeBlankGRiD
        GRiD
        ArrayOfUri
        ArrayOfStr
        ArrayOfTransformStep
        ArrayOfFile
        ArrayOfQAFlags
        ArrayRefStripUndefs
        ArrayOfState51Objects
        BiggerInt
        CSSColour
        Percentage
        PricetierCreationData
        PreorderCreationData
        WhitespaceTrimmedStr
        MaybeWhitespaceTrimmedStr
        PriceMappingResult
        SupportedDDEXVersion
        DDEXMessageType
        OrganisationTypeName
        LDAPOrganisationTypeName
        LDAPPassword
        TLA
        PackageName
        QAFlag
        PPLOwnerType
        RTTicket
        ExplicitStatus
        ImgDimensionStr
        AspectRatio
        MaybeMySQLDateTime
        InstanceID
    )];

# import builtin types
use MooseX::Types::Moose qw/ Undef /;
use MooseX::Types::Varchar qw/ Varchar /;
use MooseX::Types::DateTime::MySQL qw{MySQLDate MySQLDateTime};
use MooseX::Types::DateTime qw{DateTime};
use Moose::Util::TypeConstraints;
use state51::Validate;
use Try::Tiny;
use Math::BigInt;
use Data::Dumper;
use Digest::MD5;
use List::MoreUtils qw/ any /;
use Params::Util qw(_CLASS);

enum DDEXMessageType, [qw/
    ern:NewReleaseMessage
    FtpAcknowedgeMessage
    echo:ManifestMessage
    mlc:ManifestMessage
    mlc:DeclarationOfSoundRecordingRightsClaimMessage
/];

enum ExplicitStatus, [qw/yes no unknown cleaned/];
enum FileStatus, [qw/pending live junk deleted lost local trashed/];
enum CurrencyName, [qw/
    adp aed afa all amd ang aoa ars ats aud awg azm bam bbd bdt bef
    bgn bhd bif bmd bnd bob bov brl bsd btn bwp byr bzd cad cdf chf
    clp cny cop crc cup cve cyp czk dem djf dkk dop dzd eek egp ern
    esp etb eur fim fjd fkp frf gbp gel ghc gip gmd gnf grd gtq gwp
    gyd hkd hnl hrk htg huf idr iep ils inr iqd irr isk itl jmd jod
    jpy kes kgs khr kmf kpw krw kwd kyd kzt lak lbp lkr lrd lsl ltl
    luf lvl lyd mad mdl mgf mkd mmk mnt mop mro mtl mur mvr mwk mxn
    myr mzm nad ngn nio nlg nok npr nzd omr pab pen pgk php pkr pln
    pte pyg qar rol rub rwf sar sbd scr sdd sek sgd shp sit skk sll
    sos srd std svc syp szl thb tjs tmm tnd top tpe trl ttd twd tzs
    uah ugx usd uyu uzs veb vnd vuv wst xaf xcd xof xpf yer yum zar
    zmk zwd
/];

subtype PackageName, as Str, where { _CLASS($_) };

subtype MaybeBlankCurrencyName, as Str,
    where { $_ eq '' || is_CurrencyName($_) };

subtype Barcode, as Int,
    where { try { state51::Validate::assert_barcode($_); 1; } },
    message { local $@; eval { state51::Validate::assert_barcode($_); }; my $error = $@; $error =~ / at.+/; $error };

coerce Barcode, from Int, via { "0$_" };

subtype ISBN10, as Int,
    where { length($_) == 10 };

foreach my $type (Barcode, ISBN10) {
    coerce $type, from Str, via { # Just trim all whitespace
        s!\s+!!g; $_;
    };
}

coerce Barcode, from ISBN10, via {
    $_ = '978'.substr($_, 0, -1); # Prepend 978, Throw away last digit
    $_ .= state51::Validate::calc_mod10_check_digit($_); # Add checksum
    return $_;
};

subtype DPID, as Str,
    where {
        length($_) == 18 && try { state51::Validate::assert_ddex_party_id($_) }
    };

coerce DPID, from Str,
    via { s/-//g; s/^\s*//; s/^DDEX:\s*//i; s/^s*$//; $_; };

subtype ISNI, as Str,
    where {
        length($_) == 16 && try { state51::Validate::assert_ISNI($_) }
    };

subtype BlankStr, as Str,
    where { $_ eq '' };

subtype ISRC, as Str,
    where { $_ eq do { try { state51::Validate::assert_isrc($_) } } ? 1 : 0 };

coerce ISRC, from Str,
    via { s/[\s-]+//g; $_; };

# fail silently so the coercion can do its thing
my $validate_grid = sub { try { state51::Validate::assert_normalised_grid($_); 1 } };

my $normalise_grid = sub { s/\W//g; $_ = uc($_); };

subtype MaybeBlankGRiD, as Str,
    where { $_ eq '' || $validate_grid->($_)};

# extract like above
coerce MaybeBlankGRiD, from Str,
    via { $normalise_grid->($_); };

subtype GRiD, as Str,
    where { $validate_grid->($_); };

coerce GRiD, from Str,
    via { $normalise_grid->($_); };

subtype ArrayOfStr, as ArrayRef[Str],
    where { 1 };

subtype ArrayOfState51Objects, as ArrayRef['state51::Base'],
    where { 1 };

class_type 'state51::Uri';

subtype ArrayOfUri, as ArrayRef['state51::Uri'],
    where { 1 };

coerce ArrayOfUri, from ArrayOfStr,
    via { my @l = @$_; [
            map { state51::Uri->from_string($_) } grep { length($_) } @l
    ]};

coerce ArrayOfUri, from Str,
    via { [ length($_) ? state51::Uri->from_string($_) : () ] };

coerce ArrayOfUri, from Undef,
    via { [] };

class_type 'state51::TransformStep';

subtype ArrayOfTransformStep, as ArrayRef['state51::TransformStep'],
    where { 1 },
    message { local $Data::Dumper::Maxdepth = 1; "Not an array of state51::TransformSteps: " . Dumper($_) };;

class_type 'state51::File::Audio';

subtype ArrayOfFile, as ArrayRef['state51::File'],
    where { 1 },
    message { local $Data::Dumper::Maxdepth = 1; "Not an array of state51::Files: " . Dumper($_) };

class_type 'state51::Country';

coerce 'state51::Country', from Str,
    via { state51::Country->from_string($_) };


# BiggerInt allows us to stash a Math::BigInt into a field, and have it
# magically squashed into a string, where it should be stored intact.  It
# does not preserve the Math::BigInt'ness.

subtype BiggerInt, as Int;

class_type 'Math::BigInt';

coerce BiggerInt, from 'Math::BigInt',
    via { $_->bstr };

subtype CSSColour, as Str,
    where { /^([0-9A-F]{6}|Z{6})$/ };

coerce CSSColour, from Str,
    via { uc($_) };

coerce CSSColour, from Undef,
    via { 'ZZZZZZ' };

subtype Percentage, as Int,
    where { $_ >= 0 and $_ <= 100 };



##Used in Request::Export->set_price_tiers
subtype PricetierCreationData, as Dict
    [
        release_format => class_type('state51::ReleaseFormat'),
        price_tier => Str,
        destination => class_type('state51::OrganisationAttribute::MusicService'),
    ];

##Used in Request::Export->set_preorder_dates
subtype PreorderCreationData, as Dict
    [
        release_format => class_type('state51::ReleaseFormat'),
        date => Str,
        destination => class_type('state51::OrganisationAttribute::MusicService'),
    ];

subtype WhitespaceTrimmedStr, as Str,
    where { !/^\s/ and ! /\s$/ };
coerce WhitespaceTrimmedStr, from Undef, via { '' };

subtype MaybeWhitespaceTrimmedStr, as Maybe[Str],
    where { defined($_) ? (!/^\s/ and !/\s$/ and /\S/) : 1 };

# XXX Varchar getting upset by coercion under Strawberry Perl v5.12.3.
subtype PriceMappingResult, as Str,
    where { length($_)<=90 and !/^\s/ and ! /\s$/ };

foreach my $name (WhitespaceTrimmedStr, PriceMappingResult) {
    coerce $name, from Str,
    via { s/^\s+//; s/\s+$//; $_; };
}

coerce MaybeWhitespaceTrimmedStr, from Str,
    via { s/^\s+//; s/\s+$//; length($_) ? $_ : undef; };

subtype SupportedDDEXVersion, as enum([qw/20110222 20110307 YouTube 20160510 20190924 382SubscriptionStream 382Twitch 20220728 383Twitch /]),
    message { "DDEX version must be supported (either 20110222, 20110307, YouTube, 20160510, 20190924, 382SubscriptionStream, 382Twitch, 20220728 or 383Twitch not $_" };

subtype OrganisationTypeName,
    as enum([ 'Developer', 'Digital Music Service', 'Business Service', 'Media Provider']);

subtype LDAPOrganisationTypeName,
    as enum([ 'Developer', ' MusicService', 'BusinessService', 'MediaProvider']);

coerce LDAPOrganisationTypeName,
    from OrganisationTypeName,
    via { s/ //g; s/^Digital//; $_; };

subtype LDAPPassword,
    as Str,
    where { /^{\w+}.+==$/};

coerce LDAPPassword,
    from Str,
    via {
        my $digest = Digest::MD5->new;
        $digest->add($_);
        '{MD5}' . $digest->b64digest . '=='
    };

subtype TLA,
    as Str,
    where { /^([A-Z_0-9]{3}|)$/ };

coerce 'state51::OrganisationAttribute::MusicService', from TLA,
    via { try { state51::OrganisationAttribute::MusicService->fetch_one($_) } };

# XXX - FIXME, the enums here are also in QAFlagType.pm.. They should be moved to named types in here, then Mixin::Statusfield should be able to
#              optionally use a type name, rather than generating an anon enum from a list
subtype QAFlag,
    as Dict[
        qa_advice_id => Varchar[50],
        category => enum([qw/Metadata DPRule ArtistBlackList Video Artwork ArtistBlackListApple ArtistBlackListSpotify AudibleMagicMatch ArtistBlackListState51/]),
        severity => enum([qw/ Info Warning Critical DPCritical /]),
        is_blocker  => Optional[Bool],
    ],
    where { 1 };

subtype ArrayOfQAFlags,
    as ArrayRef[QAFlag],
    where { 1 };

subtype PPLOwnerType,
    as enum([
        "OriginalHolder",
        "Licensee",
        "Successor in title",
    ]);

subtype RTTicket, as Int,
    where { $_ >= 1 };

subtype ArrayRefStripUndefs, as ArrayRef,
    where { !any { !defined $_ } @$_ };
coerce ArrayRefStripUndefs, from ArrayRef,
    via { [ grep { defined $_ } @$_ ] };

subtype ImgDimensionStr, as Str,
    where { /^\d+x\d+$/ };

subtype AspectRatio, as Str,
    where { /^\d+(.\d+)?:\d+(.\d+)?$/ };

##

subtype MaybeMySQLDateTime, as Str,
    where { (not defined $_) or to_MySQLDateTime($_) or to_MySQLDate($_) };

coerce MaybeMySQLDateTime, from Undef, via { $_ };

coerce MaybeMySQLDateTime, from DateTime, via { to_MySQLDateTime($_) };

# this is supposed to be the "full" id, so including the class_type on the end
# TODO: validate class_type ?
subtype InstanceID, as Int,
    where { $_ >= 1 };

1;

