package state51::Types::DateApproximate;
use Moose;
use Moose::Util::TypeConstraints;

use DateTime;

# TODO: It's not really a proper moose type...

=head1 NAME

state51::Types::DateApproximate

=head1 DESCRIPTION

Allow for the handling of dates of varying exactness. This is useful when you're collecting dates for
e.g. a rights claim and the truly important element is the year but you want to allow year/month and
year/month/day where it's available. ->is_approximate is true when the date only has a year part.

Dates will be validated and normalised.

e.g.

my $date = state51::Types::DateApproximate->new('2019');

leaves $date as '2019'

my $date = state51::Types::DateApproximate->new('2019-01');

leaves $date as '2019-01'

my $date = state51::Types::DateApproximate->new('2019-01-32');

will throw

my $date = state51::Types::DateApproximate->new('2019-1-5');
print $date;

will print "2019-01-05"

=cut

has 'date' => (is => 'rw', isa => 'Str', required => 1);

has 'datetime' =>
  (is => 'ro', isa => 'DateTime', lazy => 1, builder => '_build_datetime');

has 'is_approximate' =>
  (is => 'rw', isa => 'Bool', lazy => 1, builder => '_build_is_approximate');

has 'month_is_significant' =>
  (is => 'rw', isa => 'Bool', lazy => 1, default => 0);
has 'day_is_significant' =>
  (is => 'rw', isa => 'Bool', lazy => 1, default => 0);

around BUILDARGS => sub {
    my $orig  = shift;
    my $class = shift;

    if (@_ == 1 && !ref $_[0]) {
        return $class->$orig(date => $_[0]);
    }
    else {
        return $class->$orig(@_);
    }
};

sub BUILD {
    my $self = shift;

    $self->init_object_from_date($self->date());
}

sub init_object_from_date {
    my ($self, $date) = @_;

    my ($yy, $mm, $dd) = split(/\D/, $date);
    my @params;
    if ($yy) {
        push @params, (year => $yy);
    }
    if ($mm) {
        push @params, (month => $mm);
        $self->month_is_significant(1);
    }
    if ($dd) {
        push @params, (day => $dd);
        $self->day_is_significant(1);
    }

    my $dt = DateTime->new(@params);
    return $dt;
}

sub _build_datetime {
    my ($self) = @_;

    my $dt = $self->init_object_from_date($self->date);
    return $dt;
}

sub _build_is_approximate {
    my ($self) = @_;

    my $dt = $self->init_object_from_date($self->date);
    if ($self->month_is_significant) {
        return 0;
    }

    return 1;
}

use overload
  fallback => 1,
  '""'     => 'stringify';

sub stringify {
    my ($self) = @_;

    my $dt = $self->datetime;
    if ($self->day_is_significant()) {
        return sprintf '%s-%02d-%02d', $dt->year, $dt->month, $dt->day;
    }
    elsif ($self->month_is_significant()) {
        return sprintf '%s-%02d', $dt->year, $dt->month;
    }

    return $dt->year();
}

subtype 'approx_date' => as 'state51::Types::DateApproximate';
coerce 'approx_date' => from 'Str' => via {
    if ($_) {
        state51::Types::DateApproximate->new($_);
    }
};

1;
