package state51::Validate;

use strict;
use warnings;

use Date::Calc qw(check_date check_time);
use Email::Valid::Loose;
use Carp qw/ confess /;

use state51::ValidateJSBridge;

=head1 NAME

state51::Validate - module that exports lots of validation routines

=head1 DESCRIPTION

creates a lot of validation routines

=cut

sub assert_email {
    my ($value) = @_;

    die('InvalidEmail')
      unless
        $value;

    die('InvalidEmail')
      unless
        Email::Valid::Loose->address($value);
    return($value);
}

##

sub assert_nonempty_string {
    my ($value) = @_;

    die('InvalidStringEmpty')
      unless
        (defined($value) and length($value));

    return($value);
}

##

sub assert_barcode {
    my ($value) = @_;

    my $validator = state51::ValidateJSBridge->instance();
    return $validator->assert_barcode($value);
}

sub assert_grid {
    my ($value) = @_;
    $value =~ s/-//g;

    return assert_normalised_grid($value);
}

# http://www.ifpi.org/content/section_resources/grid-standard.html#Construction
sub assert_normalised_grid {
    my ($value) = @_;

    die("Empty\n")
      unless
        $value;

    die("Wrong Length\n")
      unless
    length($value) == 18;

    $value = uc($value);
    my $orig_check = substr($value, 17, 1);
    $value = substr($value, 0, 17);

    my $calc_check = calc_mod36_37_check_digit($value);

    if ($calc_check ne $orig_check) {
        die('InvalidGRIDIncorrectCheckSum');
    }

    return $value.$orig_check;
}

# value is the barcode without a checkdigit
sub calc_mod10_check_digit {
    my ($value) = @_;

    my $calc_checksum = 0;

    my $mult;
    if (length($value) % 2) {
        $mult = 3;
    }
    else { # for EAN-13 the first digit is "even"
        $mult = 1;
    }

    foreach my $char (split('',$value)) {
        $calc_checksum += ($char * $mult);

        if ($mult == 1) {
            $mult = 3;
        }
        else {
            $mult = 1;
        }
    }

    # The check digit is the number which, when added to the totals calculated
    # results in a number evenly divisible by 10.
    $calc_checksum = 10 - ($calc_checksum % 10);
    $calc_checksum = 0 if ($calc_checksum == 10);

    return $calc_checksum;
}

sub g_value { # from char to value
    my ($char) = @_;
    my $value = ord($char) - 55;
    $value += 7 if $value < 10;
    return $value;
}

sub g_char { # from value to char
    my ($value) = @_;
    $value -= 7 if $value < 10;
    my $char = chr($value + 55);
    return $char;
}

# see https://www.google.co.uk/url?sa=t&rct=j&q=&esrc=s&source=web&cd=4&ved=0ahUKEwi-mpG406bUAhXKDMAKHZdECMAQFgg0MAM&url=http%3A%2F%2Fhsevi.ir%2FRI_Standard%2FFile%2F7794&usg=AFQjCNEgYhWH6Qz_jFCu06GHDyb7HrC0HQ&cad=rja
sub calc_mod36_37_check_digit { # ISO 7064, mod 37,36
    my ($value) = @_; # grid/dpid without digit, 17 chars

    my $p = 36;
    foreach my $char (split('',$value)) {
        my $s = ($p % 37) + g_value($char);
        $p = ($s % 36);
        $p = 36 if ($p == 0);
        $p = $p*2;
    }

    my $check_digit = 0;
    my $not_found = 1;
    while ($check_digit <= 35 and $not_found) {
        if (((($p % 37) + $check_digit) % 36) == 1) {
            $not_found = 0;
        }
        $check_digit++;
    }

    return g_char($check_digit-1);
}

sub calc_mod11_2_check_digit { # ISO 7064 mod11-2
    my ($value) = @_; # value lacks check digit

    my $p = 0;
    for (my $index = 0; $index < length($value); $index++) {
        my $char = substr($value, $index, 1);
        $p = 2 * ($p + $char);
    }
    $p %= 11;
    my $result = (12 - $p) % 11;

    if ($result == 10) { $result = 'X' }

    return $result;
}

sub assert_ddex_party_id { # https://kb.ddex.net/display/HBK/DDEX%2BParty%2BIdentifier%2BStandard
    my ($value) = @_;

    die('InvalidDPIDEmpty')
        unless
            $value;

    $value =~ s/-//g;
    $value =~ s/^DDEX://i;

    die('InvalidDPIDWrongLength')
        unless
            length($value) == 18;

    $value = uc($value);
    my $orig_check = substr($value, 17, 1);
    $value = substr($value, 0, 17);

    my $calc_check = calc_mod36_37_check_digit($value);

    if ($calc_check ne $orig_check) {
        die("InvalidDPIDIncorrectCheckSum");
    }

    return $value.$orig_check;
}

sub assert_ISNI { # https://www.ifla.org/files/assets/uca/unimarc_updates/AUTHORITIES/a_010_new2016.pdf
    my ($value) = @_;

    my $validator = state51::ValidateJSBridge->instance();
    return $validator->assert_isni($value);
}

# http://www.ifpi.org/isrc/isrc_handbook.html
# 2 characters (iso2code with caveats), 3 characters, 2 digits(year), 5 digits

#
# as a side effect, it returns a normalised isrc
sub assert_isrc {
    my $isrc = shift;

    my $validator = state51::ValidateJSBridge->instance();
    return $validator->assert_isrc($isrc);
}

##
# FIXME - This regex / date code is horrible!!!
#         These are MooseX::Types::DateTime::MySQL timestamps, are they not? So we should use
#         that?
sub assert_preorder_date {
    my ($preorder_date, $release_date) = @_;
    confess("No preorder_date (yyyy-mm-dd 00:00:00) given") unless defined $preorder_date;
    confess("No release_date given") unless defined $release_date;

    return $preorder_date if ($preorder_date eq '0000-00-00 00:00:00');

    my ($year, $mon, $mday, $hour, $minute, $second) = $preorder_date =~ /^(\d+)-(\d+)-(\d+)(?: (\d+):(\d+):(\d+))?$/;

    confess "Date '$preorder_date' is not valid (yyyy-mm-dd hh:mm:ss)"
        unless check_date($year,$mon,$mday);

    if (defined $hour) {
        confess "Date '$preorder_date' is not valid (yyyy-mm-dd hh:mm:ss)"
            unless check_time($hour, $minute, $second);
    }
    else {
        $preorder_date .= " 00:00:00";
    }

    die("Pre Order Date ($preorder_date) must be before release date ($release_date)") if ($preorder_date ge $release_date);

    return $preorder_date;
}

END {
    my $validator = state51::ValidateJSBridge->instance();
    $validator->clear_js_instance();
}

##
1;
