package state51::ValidateJSBridge;

use Moose;
use MooseX::Singleton;
use JavaScript::QuickJS;
use File::ShareDir qw/ module_dir /;

# This module uses QuickJS to compile openimp_validate
# and build callbacks to allow perl to call functions there as if they were native
# this allows us to share validation code between the front and back ends

has 'js_instance' => (
    is      => 'ro',
    isa     => 'JavaScript::QuickJS',
    builder => '_build_js_instance',
    lazy    => 1,
    clearer => '_clear_js_instance',
);

has 'validators' => (
    is      => 'ro',
    isa     => 'HashRef',
    builder => '_build_validators',
    lazy    => 1,
    clearer => '_clear_validators',
);

sub _build_js_instance {
    # this allows you to call console.log in the js
    # return JavaScript::QuickJS->new()->helpers();
    return JavaScript::QuickJS->new();
}

sub _build_validators {
    my $self = shift;

    my $js_file = module_dir(__PACKAGE__)."/openimp_validate.js";

    my $retval;
    # this is a bit of magic that lets return a value despite eval_module
    $self->js_instance->set_globals(_return => sub { $retval = shift });

    my $js_code = <<~END;
        import { assert_isrc, assert_ISNI, assert_barcode } from '$js_file';
        _return({
            assert_isrc,
            assert_ISNI,
            assert_barcode
        });
        END

    # eval_module returns a promise
    # and we are returning functions in a dispatch table we can use as callbacks in perl
    $self->js_instance->eval_module($js_code);

    return $retval;
}

sub assert_isrc {
    my ($self, $isrc) = @_;
    my $validator = $self->validators->{assert_isrc};
    return $validator->($isrc);
}

sub assert_isni {
    my ($self, $isni) = @_;
    my $validator = $self->validators->{assert_ISNI};
    return $validator->($isni);
}

sub assert_barcode {
    my ($self, $barcode) = @_;
    my $validator = $self->validators->{assert_barcode};
    return $validator->($barcode);
}

# DEMOLISH doesn't seem to work here, because the order things are
# destroyed matters, so we have this explicit cleanup method
# and Validate calls it inside END
sub clear_js_instance {
    my $self = shift;

    $self->validators->{assert_isrc} = undef;
    $self->validators->{assert_ISNI} = undef;
    $self->validators->{assert_barcode} = undef;
    $self->_clear_validators();
    $self->_clear_js_instance();
}

__PACKAGE__->meta->make_immutable;

1;
