#!/usr/bin/env perl

use strict;
use warnings;

BEGIN { $ENV{PLATFORM_MQ_SERVER} ||= 'localhost' }

use state51::RabbitMQ;
use Test::More;
use Test::More skip_all => "Test runs forever";
use state51::AnyEvent qw/ run_loop_till_idle /;
use FindBin;
use File::Spec;
use Try::Tiny;
use lib File::Spec->catdir("$FindBin::Bin", "lib");

use TestJob;
use state51::Shared::ForkExecPerl qw/ run_perl /;
use File::Temp qw/ tempdir /;
use LWP::Simple qw/ get /;
use JSON qw/ decode_json /;


my $exchange_name = TestJob->exchange_name();

my $mq = state51::RabbitMQ->instance->mq();
isa_ok($mq, 'AnyEvent::RabbitMQ');

my $j = AnyEvent->condvar;

isa_ok($j, 'AnyEvent::CondVar');

my $channel;

my $channel_close_ok = 0;
$mq->open_channel(
    queue => $exchange_name,
    on_success => sub {
        $channel = shift;
        $j->send();
    },
    on_close => sub { fail("Closed channel") unless $channel_close_ok },
    on_failure => sub { warn "didn't work"; $j->send(); },
);

$j->recv();
isa_ok($channel, 'AnyEvent::RabbitMQ::Channel');

$j = AnyEvent->condvar;
diag "Declaring exchange";

$j = AnyEvent->condvar;
$channel->declare_exchange(
    exchange => $exchange_name,
    type => "topic",
    durable => 1,
    on_success => sub {
        $j->send(@_);
    },
    on_failure => sub { warn "didn't work"; $j->send(@_); },
);

diag "waiting for response";
diag $j->recv();

ok 1, "opened channel and exchange";

$j = AnyEvent->condvar;
$channel->declare_queue(
    durable => 1,
    queue => $exchange_name,
    on_success => sub { $j->send(@_); },
    on_failure => sub { warn "didn't work"; $j->send(@_); },
);
diag $j->recv();

$j = AnyEvent->condvar;
$channel->bind_queue(
    queue => $exchange_name,
    exchange => $exchange_name,
    routing_key => "foo",
    on_success => sub { $j->send(@_); },
    on_failure => sub { warn "didn't work"; $j->send(@_); },
);
diag $j->recv();

my $td = tempdir( CLEANUP => 1 );

my $job = TestJob->new( file_to_poke => File::Spec->catdir($td, "hallo") );
diag $job->freeze();

$j = AnyEvent->condvar;

$channel->purge_queue(
    queue => $exchange_name,
    on_success => sub { $j->send(); },
    on_failure => sub { $j->send(); },
);
$j->recv();

$channel->publish(
    body => $job->freeze(),
    exchange => $exchange_name,
    routing_key => "foo",
);


my $keep_running_file = File::Spec->catdir($td, "keep_alive");
open(my $keep_running_file_fh, ">", $keep_running_file) or die $!;
close $keep_running_file_fh;

my $pid = run_perl(
    "state51::Script::MessageJobs::SharedWorkerPools",
    "state51::Script::MessageJobs::SharedWorkerPools->new_with_options->run",
    "--pools", "1:decoy",
    "--pools", "2:$exchange_name",
    "--pools", "3:decoy2",
    "--keep_running_file", $keep_running_file
);

while (! -e $job->file_to_poke) {
    run_loop_till_idle();
    diag "now waiting for job to run from $$ ".$job->file_to_poke;
    sleep 1;
}

{
    open(my $fh, "<", $job->file_to_poke) or die $!;
    is_deeply([ <$fh> ], ["HELLO WORLD!\n"]);
}

# Now we create a load of jobs, and kill almost immediately, so that we can
# watch the workers terminate.

for my $i (1..30) {
    my $f = TestJob->new( file_to_poke => File::Spec->catdir($td, $i), wait_for => 3 )->freeze();
    $channel->publish(
        body => $f,
        exchange => $exchange_name,
        routing_key => "foo",
    );
}

my $many_job_start = time;

my $waiting_for_first = File::Spec->catdir($td, 1);

while (! -e $waiting_for_first) {
    run_loop_till_idle();
    diag "now waiting for first out of many jobs to run from $$ $waiting_for_first";
    sleep 1;
}

ok((time-$many_job_start) < 30, "New events picked up instantly, no need to wait until periodic automatic run");

is(unlink($keep_running_file), 1);

# waitpid can fail on a win32 process which has already exited.  There are no zombies under Windows.
is(waitpid($pid, 0), $pid);

# Wait for workers to finish.
sleep 5;

my $total_count;
for my $i (1..30) {
    $total_count++ if (-e File::Spec->catdir($td, $i));
}

diag "$total_count jobs have been done";

ok $total_count >  0;
ok $total_count < 30;

$channel_close_ok = 1;

done_testing();

